"""Various useful functions used throughout the package"""

import math
from copy import copy


def is_list_or_tuple(var):
    """Checks if an object is a list or a tuple"""
    return isinstance(var, list) or isinstance(var, tuple)


def assert_list_or_tuple(var):
    """Asserts that an object is a list or a tuple"""
    if not is_list_or_tuple(var):
        raise TypeError(f"variable should be a list or a tuple, not a {type(var)}")


def assert_str(var):
    """Asserts that an object is a str"""
    if not isinstance(var, str):
        raise TypeError(f"variable should be a str, not a {type(var)}")


def simplify(obj):
    """Tries to simplify an object"""
    if hasattr(obj, "simplify"):
        return obj.simplify()
    return copy(obj)


def expand(obj):
    """Tries to expand an object"""
    if hasattr(obj, "expand"):
        return obj.expand()
    return copy(obj)


def is_zero(obj):
    """Tries to check if an object is considered zero"""
    if isinstance(obj, int) or isinstance(obj, float):
        return math.isclose(obj, 0, abs_tol=1e-16)
    if hasattr(obj, "is_zero"):
        return obj.is_zero()
    return obj == 0


def is_one(obj):
    """Tries to check if an object is considered one"""
    if isinstance(obj, int) or isinstance(obj, float):
        return math.isclose(obj, 1, abs_tol=1e-16)
    if hasattr(obj, "is_one"):
        return obj.is_one()
    return obj == 1


def replace_var(obj, old_variable=None, new_variable=None):
    """Tries to replace a variable in an object."""
    if hasattr(obj, "replace_var"):
        if old_variable is None:
            new_obj = copy(obj)
            for old_variable in get_variables(obj):
                new_obj = replace_var(new_obj, old_variable=old_variable)
            return new_obj
        if new_variable is None:
            new_variable = old_variable + "'"
        return obj.replace_var(old_variable, new_variable)
    return copy(obj)


def get_variables(obj):
    """Tries to get variables in an object."""
    if hasattr(obj, "get_variables"):
        return obj.get_variables()
    return set([])


def has_variable(obj, variable):
    """Tries to check if an object has a variable."""
    if hasattr(obj, "has_variable"):
        return obj.has_variable(variable)
    return False



r"""
Module for representing quantum states.

The :class:`~.BaseState`-class is intended to be subclassed for different types of states,
e.g. qubit states, qudit states, fock states etc.

The :class:`~.State`-class is then a sum of :class:`~.BaseState`s and is not
intended to be subclassed but rather agnostic to the exact :class:`~.BaseState`-class
in the sum.
"""

import abc
from collections import defaultdict

from qualg.scalars import is_scalar
from qualg.toolbox import assert_list_or_tuple, simplify, replace_var, get_variables, is_zero


class BaseState(abc.ABC):
    r"""Abstract base state class.
    Represents a single term of state and used to define how inner product should be taken etc.

    Meant to be subclassed.
    """
    @abc.abstractmethod
    def __eq__(self, other):
        pass

    @abc.abstractmethod
    def __hash__(self):
        pass

    def __matmul__(self, other):
        return self.tensor_product(other)

    @abc.abstractmethod
    def inner_product(self, other):
        """
        Takes the inner product with another :class:`~.BaseState`.

        Parameters
        ----------
        other : :class:`.BaseState`
            The right hand side of the inner product.

        Returns
        -------
        :class`~.scalar.Scalar`
            The inner product
        """
        pass

    @abc.abstractmethod
    def tensor_product(self, other):
        """
        Takes the tensor product with another :class:`~.BaseState`.

        Parameters
        ----------
        other : :class:`.BaseState`
            The right hand side of the tensor product.

        Returns
        -------
        :class:`~.BaseState`
            The tensor product
        """
        pass

    @abc.abstractmethod
    def _compatible(self, other):
        """Used to check if two states are compatible.

        For example if they have the same number of qubits.
        """
        pass

    def to_state(self):
        """Converts the base state to a state with a single term."""
        return State([self])

    @property
    def shape(self):
        """Returns the shape of the state, e.q. (2,) for a qubit.

        `None` means that the shape is undefined, e.g. if the state is infinite-dimensional.
        """
        return None

    @abc.abstractmethod
    def _bra_str(self):
        pass

    def _vector_index(self):
        """Specifies the index in an actual vector.

        `None` means that the index is undefined, e.g. if the state is infinite-dimensional.
        """
        return None


class State:
    def __init__(self, base_states=None, scalars=None):
        """A quantum state.
        Constructed as a sum of (subclass) :class:`~.BaseState`.

        Parameters
        ----------
        base_states : None or list of :class:`~.BaseState`
            The base states that sums up to this state.
            If `None`, then the state is "zero", i.e. no terms.
        scalar : None or list of :class:`~.scalar.Scalar`
            The amplitudes used when taking the sum of base states.
            If `None`, then all operators have amplitude 1.
        """
        if base_states is None:
            self._terms = defaultdict(int)
            return
        assert_list_or_tuple(base_states)
        if scalars is None:
            scalars = [1] * len(base_states)
        else:
            assert_list_or_tuple(scalars)
            if len(base_states) != len(scalars):
                raise ValueError("number of base_states and scalars must be equal")

        # NOTE assuming that all scalars can add int()
        self._terms = defaultdict(int)
        for base_state, scalar in zip(base_states, scalars):
            if not isinstance(base_state, BaseState):
                raise TypeError(f"base_states needs to be of class BaseState, not {type(base_state)}")
            if not is_scalar(scalar):
                raise TypeError(f"scalars needs to be of class Scalar, not {type(scalar)}")
            self._terms[base_state] += scalar

        # Check that all states are pairwise compatible
        base_states = list(self._terms.keys())
        for i, bs1 in enumerate(base_states):
            for j in range(i + 1, len(base_states)):
                bs2 = base_states[j]
                if not bs1._compatible(bs2):
                    raise ValueError(f"States {bs1} and {bs2} are not compatible terms")

    def _key(self):
        return set((base_state, scalar) for base_state, scalar in self._terms.items())

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._key() == other._key()

    def __hash__(self):
        return hash(self._key())

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise NotImplementedError(f"addition is not implemented for {type(other)}")
        # Check that the states are compatible
        # NOTE we only need to check the first terms of the two states.
        if not self._compatible(other):
            raise ValueError(f"other ({other}) is not compatible with self ({self})")
        new_state = State([])
        # Add this terms
        for base_state, scalar in self._terms.items():
            new_state._terms[base_state] = scalar
        # Add the other terms
        for base_state, scalar in other._terms.items():
            new_state._terms[base_state] += scalar

        # Check if there are any zero terms
        new_state._prune_zero_terms()

        return new_state

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return self + (-1 * other)

    def __rsub__(self, other):
        return other + (-1 * self)

    def __mul__(self, other):
        if not is_scalar(other):
            return NotImplemented
        new_state = State([])
        for base_state, scalar in self._terms.items():
            new_state._terms[base_state] = scalar * other

        return new_state

    def __rmul__(self, other):
        return self * other

    def __str__(self):
        to_return = ""
        for base_state, scalar in self._terms.items():
            to_return += f"{scalar}*{base_state} + "
        return to_return[:-3]

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(list(self._terms.items()))})"

    def __len__(self):
        return len(self._terms)

    def __matmul__(self, other):
        return self.tensor_product(other)

    def __iter__(self):
        return iter(self._terms.items())

    def get_scalar(self, base_state):
        """Returns the scalar of the given base_state"""
        return self._terms.get(base_state, 0)

    def inner_product(self, other, first_replace_var=True):
        """
        Takes the inner product with another :class:`~.State`.

        Parameters
        ----------
        other : :class:`.State`
            The right hand side of the inner product.
        first_replace_var (optional) : bool
            Whether to replace all varibles of the right hand side with
            new ones (only done for `SingleVarFunctionScalar`).
            This can be useful when the two states are actually integrals
            over the variables and should therefore be different.

        Returns
        -------
        :class`~.scalar.Scalar`
            The inner product
        """
        if not isinstance(other, self.__class__):
            raise NotImplementedError(f"inner product is not implemented for {type(other)}")
        if not self._compatible(other):
            raise ValueError(f"other ({other}) is not compatible with self ({self})")
        # NOTE if BaseState are assumed to be orthogonal be don't need to do the
        # product of base states.
        if first_replace_var:
            other = replace_var(other)
        inner = 0
        for self_base_state, self_scalar in self._terms.items():
            for other_base_state, other_scalar in other._terms.items():
                factors = [
                    self_scalar.conjugate(),
                    other_scalar,
                    self_base_state.inner_product(other_base_state),
                ]
                if any(is_zero(factor) for factor in factors):
                    continue
                inner += (self_scalar.conjugate() * other_scalar) * self_base_state.inner_product(other_base_state)

        return inner

    def tensor_product(self, other):
        """
        Takes the tensor product with another :class:`~.State`.

        Parameters
        ----------
        other : :class:`.State`
            The right hand side of the tensor product.

        Returns
        -------
        :class:`~.State`
            The tensor product
        """
        if not isinstance(other, self.__class__):
            raise NotImplementedError(f"tensor product is not implemented for {type(other)}")
        tensor = State()
        for self_base_state, self_scalar in self._terms.items():
            for other_base_state, other_scalar in other._terms.items():
                tensor += State(base_states=[self_base_state.tensor_product(other_base_state)],
                                scalars=[self_scalar * other_scalar])

        return tensor

    def simplify(self):
        """Tries to simplify the state."""
        new_state = State()
        for base_state, scalar in self._terms.items():
            new_state._terms[base_state] = simplify(scalar)

        new_state._prune_zero_terms()

        return new_state

    def replace_var(self, old_variable, new_variable):
        """
        Replaces a variable with another.
        """
        new_state = State()
        for base_state, scalar in self._terms.items():
            new_base_state = replace_var(base_state, old_variable, new_variable)
            new_scalar = replace_var(scalar, old_variable, new_variable)
            new_state._terms[new_base_state] = new_scalar

        return new_state

    def get_variables(self):
        """
        Returns the variable of this operator.
        """
        vars = set([])
        for term in self:
            for part in term:
                vars |= get_variables(part)

        return vars

    def _compatible(self, other):
        if not isinstance(other, State):
            return False
        if len(self) == 0 or len(other) == 0:
            return True

        # NOTE we only need to check one of the terms
        self_term = next(iter(self._terms.keys()))
        other_term = next(iter(other._terms.keys()))
        return self_term._compatible(other_term)

    def _prune_zero_terms(self):
        to_remove = []
        for base_state, scalar in list(self._terms.items()):
            if scalar == 0:
                to_remove.append(base_state)

        for base_state in to_remove:
            self._terms.pop(base_state)

    def _bra_str(self):
        to_return = ""
        for base_state, scalar in self._terms.items():
            to_return += f"{scalar.conjugate()}*{base_state._bra_str()} + "
        return to_return[:-3]



r"""
Module for representing operators of quantum states.

An :class:`~.Operator` is constructed as a sum of :class:`~.BaseOperator`s, which define how
the terms of  the operator act on :class:`~.states.BaseState` and in turn :class:`~.states.State`

The :class:`~.BaseOperator`-class represents a single term of the form \|left><right\|,
where left and right are :class:`~.states.BaseState`'s. The :class:`~.BaseOperator` is
agnostic to which subclass of :class:`~.states.BaseState` is used.

The :class:`~.Operator`-class is then a sum of :class:`~.BaseOperator`.
"""

import numpy as np
from collections import defaultdict

from qualg.scalars import is_scalar, is_number
from qualg.states import BaseState, State
from qualg.toolbox import assert_list_or_tuple, simplify, replace_var, get_variables, is_zero
from qualg.integrate import integrate


class BaseOperator:
    def __init__(self, left, right):
        r"""Represents a single term of an operator, i.e. \|left><right\|,
        where left and right are :class:`~.states.BaseState`'s.

        Parameters
        ----------
        left : :class:`~.states.BaseState`
            Left side of operator.
        right : :class:`~.states.BaseState`
            Right side of operator.
        """
        if not all(isinstance(s, BaseState) for s in [left, right]):
            raise TypeError(f"Both left and right should be of type BaseState, not {type(left)} or {type(right)}")
        self._left = left
        self._right = right

    def _key(self):
        return (self._left, self._right)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._key() == other._key()

    def __hash__(self):
        return hash(self._key())

    def __mul__(self, other):
        if not isinstance(other, State):
            return NotImplemented
        if not self._mul_compatible(other):
            raise TypeError(f"other ({other}) is not multiplication compatible with self ({self})")
        new_state = State([])
        for base_state, scalar in other._terms.items():
            new_state._terms[self._left] += self._right.inner_product(base_state) * scalar

        return new_state

    def __str__(self):
        return f"Op[{self._left}{self._right._bra_str()}]"

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._left)}, {repr(self._right)})"

    @property
    def shape(self):
        """Returns the shape of the operator, e.q. (2, 2) for a single-qubit operator.

        `None` means that the shape is undefined, e.g. if the state is infinite-dimensional.
        """
        return (self._left.shape[0], self._right.shape[0])

    def _mul_compatible(self, other):
        """Used to check if an operator or state is compatible for multiplication.

        For example if they act on the same number of qubits.
        """
        if isinstance(other, self.__class__):
            return self._right._compatible(other._left)
        elif isinstance(other, State):
            return self._right._compatible(next(iter(other._terms)))
        elif isinstance(other, BaseState):
            return self._right._compatible(other)
        return False

    def _add_compatible(self, other):
        """Used to check if an operator is compatible for addition.

        For example if they act on the same number of qubits.
        """
        if not isinstance(other, self.__class__):
            return False
        return self._left._compatible(other._left) and self._right._compatible(other._right)

    def to_operator(self):
        """Converts the base operator to an :class:`~.Operator` with a single term."""
        return Operator([self])

    def replace_var(self, old_variable, new_variable):
        """
        Replaces a variable with another.
        """
        return BaseOperator(replace_var(self._left, old_variable, new_variable),
                            replace_var(self._right, old_variable, new_variable))

    def get_variables(self):
        """
        Returns the variable of this operator.
        """
        vars = set([])
        for bs in [self._left, self._right]:
            vars |= bs.get_variables()

        return vars

    def _matrix_index(self):
        return self._left._vector_index(), self._right._vector_index()


class Operator:
    def __init__(self, base_ops=None, scalars=None):
        """
        An operator represented as a sum of :class:`~.BaseOperator` of a subclass thereof.

        Parameters
        ----------
        base_ops : None or list of :class:`~.BaseOperator`
            The base operators that sums up to this operator.
            If `None`, then the operator is "zero", i.e. no terms.
        scalar : None or list of :class:`~.scalar.Scalar`
            The amplitudes used when taking the sum of base operators.
            If `None`, then all operators have amplitude 1.
        """
        # NOTE assuming that all scalars can add int()
        self._terms = defaultdict(int)
        if base_ops is None:
            return
        assert_list_or_tuple(base_ops)
        if scalars is None:
            scalars = [1] * len(base_ops)
        else:
            assert_list_or_tuple(scalars)
            if len(base_ops) != len(scalars):
                raise ValueError(f"number of base operators ({len(base_ops)})"
                                 f"and scalars ({len(scalars)}) are not equal")

        for op_term, scalar in zip(base_ops, scalars):
            if not isinstance(op_term, BaseOperator):
                raise TypeError(f"elements of base_ops should be instances of BaseOperator, not {type(op_term)}")
            if not is_scalar(scalar):
                raise TypeError(f"elements of scalars should be instances of Scalar, not {type(op_term)}")
            self._terms[op_term] += scalar

        base_ops = list(self._terms.keys())
        # Check that all base_ops are pairwise compatible
        for i, bo1 in enumerate(base_ops):
            for j in range(i + 1, len(base_ops)):
                bo2 = base_ops[j]
                if not bo1._add_compatible(bo2):
                    raise ValueError(f"Base operators {bo1} and {bo2} are not compatible terms")

    def _key(self):
        return set((base_op, scalar) for base_op, scalar in self._terms.items())

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._key() == other._key()

    def __hash__(self):
        return hash(self._key())

    def __mul__(self, other):
        if not self._mul_compatible(other):
            raise ValueError(f"operator not multiplication compatible with {other}")
        if is_scalar(other):
            return self._mul_scalar(other)
        elif isinstance(other, State):
            return self._mul_state(other)
        elif isinstance(other, Operator):
            return self._mul_operator(other)
        return NotImplemented

    def _mul_state(self, state):
        # compute output state for each term of the operator
        new_state = State([])
        for base_op, scalar in self._terms.items():
            if is_zero(scalar):
                continue
            new_state += scalar * (base_op * state)

        return new_state

    def _mul_scalar(self, scalar):
        new_op = Operator([])
        for base_op, c in self._terms.items():
            new_op._terms[base_op] += scalar * c

        return new_op

    def _mul_operator(self, operator):
        new_op = Operator()
        for self_base_op, self_scalar in self._terms.items():
            for other_base_op, other_scalar in operator._terms.items():
                new_base_op = BaseOperator(self_base_op._left, other_base_op._right)
                new_scalar = self_base_op._right.inner_product(other_base_op._left) * self_scalar * other_scalar

                # Integrate out variables which are not in base operator
                # TODO, should this be optional?
                scalar_variables = get_variables(new_scalar) - get_variables(new_base_op)
                new_scalar = integrate(new_scalar, scalar_variables)

                if is_zero(new_scalar):
                    continue
                new_op._terms[new_base_op] += new_scalar

        return new_op

    def __add__(self, other):
        if not self._add_compatible(other):
            raise ValueError(f"operator not addition compatible with {other}")
        # Do add
        new_op = Operator([])
        for op in [self, other]:
            for base_op, scalar in op._terms.items():
                new_op._terms[base_op] += scalar

        # Check for zero terms
        new_op._prune_zero_terms()

        return new_op

    def __radd__(self, other):
        return self + other

    def __len__(self):
        return len(self._terms)

    def __str__(self):
        to_print = ""
        for base_op, scalar in self._terms.items():
            to_print += f"{scalar}*{base_op} + "
        return to_print[:-3]

    def __repr__(self):
        base_ops = list(self._terms.keys())
        scalars = [self._terms[base_op] for base_op in base_ops]
        return f"{self.__class__.__name__}({repr(base_ops)}, {repr(scalars)})"

    def __iter__(self):
        return iter(self._terms.items())

    @property
    def shape(self):
        """Returns the shape of the operator, e.q. (2, 2) for a single-qubit operator.

        `None` means that the shape is undefined, e.g. if the state is infinite-dimensional.
        """
        if len(self) == 0:
            return (0, 0)
        else:
            return next(iter(self._terms)).shape

    def get_scalar(self, base_op):
        """Returns the scalar of the given base_op"""
        return self._terms.get(base_op, 0)

    def dagger(self):
        """
        Complex conjugate of the operator.
        """
        new_op = Operator()
        for base_op, scalar in self._terms.items():
            new_base_op = BaseOperator(left=base_op._right, right=base_op._left)
            new_scalar = scalar.conjugate()
            new_op._terms[new_base_op] += new_scalar
        return new_op

    def simplify(self):
        """
        Tries to simplify the operator, returning a new one.
        """
        new_op = Operator()
        for base_op, scalar in self._terms.items():
            new_op._terms[base_op] = simplify(scalar)

        new_op._prune_zero_terms()

        return new_op

    def replace_var(self, old_variable, new_variable):
        """
        Replaces a variable with another.
        """
        new_op = Operator()
        for base_op, scalar in self._terms.items():
            new_base_op = replace_var(base_op, old_variable, new_variable)
            new_scalar = replace_var(scalar, old_variable, new_variable)
            new_op._terms[new_base_op] = new_scalar

        return new_op

    def get_variables(self):
        """
        Returns the variable of this operator.
        """
        vars = set([])
        for term in self:
            for part in term:
                vars |= get_variables(part)

        return vars

    def to_numpy_matrix(self, convert_scalars=None, **kwargs):
        """Converts the operator to a numpy matrix.

        If there are non-number scalars then the provided function `convert_scalars`
        is used to convert a non-number scalar to a number.
        This function should then take a scalar (plus optional keyword arguments) and return a number.

        Parameters
        ----------
        convert_scalars : function
            Function to convert a non-number scalar to a number.
        **kwargs:
            Keyword-arguments to be passed to `convert_scalars`.

        Returns
        -------
        :class:`numpy.ndarray`
            The operator in numerical matrix form.
        """
        matrix = np.zeros(self.shape)
        for base_op, scalar in self:
            index = base_op._matrix_index()
            if not is_number(scalar):
                if convert_scalars is None:
                    raise ValueError("If the operator contains non-numbers, "
                                     "the function `convert_scalars` needs to be provided")
                scalar = convert_scalars(scalar, **kwargs)
            matrix[index] = scalar

        return matrix

    def _prune_zero_terms(self):
        to_remove = []
        for base_op, scalar in list(self._terms.items()):
            if scalar == 0:
                to_remove.append(base_op)

        for base_state in to_remove:
            self._terms.pop(base_state)

    def _mul_compatible(self, other):
        """Used to check if an operator or state is compatible for multiplication.

        For example if they act on the same number of qubits.
        """
        if isinstance(other, self.__class__) or isinstance(other, State):
            if len(self) == 0 or len(other) == 0:
                return True

            # NOTE we only need to check one of the terms
            self_term = next(iter(self._terms.keys()))
            other_term = next(iter(other._terms.keys()))
            return self_term._mul_compatible(other_term)
        elif is_scalar(other):
            return True
        return False

    def _add_compatible(self, other):
        """Used to check if an operator is compatible for addition.

        For example if they act on the same number of qubits.
        """
        if not isinstance(other, self.__class__):
            return False
        if len(self) == 0 or len(other) == 0:
            return True

        # NOTE we only need to check one of the terms
        self_term = next(iter(self._terms.keys()))
        other_term = next(iter(other._terms.keys()))
        return self_term._add_compatible(other_term)


def outer_product(left, right):
    r"""Creates an opertor based on the outer product of left and right, i.e. \|left><right\|.

    Parameters
    ----------
    left : :class:`~.states.BaseState`
        Left side of operator.
    right : :class:`~.states.BaseState`
        Right side of operator.

    Returns
    -------
    :class:`~.Operator`
        The new operator.
    """
    scalars = []
    base_ops = []
    for l_base_state, l_scalar in left._terms.items():
        for r_base_state, r_scalar in right._terms.items():
            scalars.append(l_scalar * r_scalar.conjugate())
            base_ops.append(BaseOperator(l_base_state, r_base_state))

    return Operator(base_ops, scalars)



"""
Containing classes for various scalars.
Used for example as output when taking inner-product of states.
In particular when the output is not a number but a symbolic variable of function.
"""

import abc
import math
from copy import copy
from collections import defaultdict
from itertools import product

from qualg.toolbox import assert_list_or_tuple, assert_str, expand, simplify, is_zero, replace_var, is_one,\
    get_variables, has_variable


def is_number(n):
    """Check if something is a number (int, float or complex)"""
    return any(isinstance(n, tp) for tp in [int, float, complex])


def is_scalar(n):
    """Check if something is considered a scalar (number of :class:`~.Scalar`)"""
    return is_number(n) or isinstance(n, Scalar)


class Scalar(abc.ABC):
    """
    Base-class for all scalars.

    Meant to be subclassed.
    """
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._key() == other._key()

    def __hash__(self):
        return hash(self._key())

    def __add__(self, other):
        # Check if one is zero
        if is_zero(other):
            return copy(self)
        if is_zero(self):
            return copy(other)

        if not is_scalar(other):
            return NotImplemented
        if isinstance(other, SumOfScalars):
            return other + self
        # Check if multiples
        self_multiple, self_scalar = _get_multiple_of_scalar(self)
        other_multiple, other_scalar = _get_multiple_of_scalar(other)
        if self_scalar == other_scalar:
            return copy(self_scalar) * (self_multiple + other_multiple)

        return SumOfScalars([self, other])

    def __mul__(self, other):
        if is_zero(other):
            return 0
        if is_one(other):
            return copy(self)
        if not is_scalar(other):
            return NotImplemented
        if isinstance(other, ProductOfScalars):
            return other * self
        return ProductOfScalars([self, other])

    def __rmul__(self, other):
        return self * other

    def __radd__(self, other):
        return self + other

    def __copy__(self):
        return eval(repr(self))
        pass

    def has_variable(self, variable):
        """Checks if scalar depends on a given variable."""
        return False

    def get_variables(self):
        """
        Returns the variable of this operator.
        """
        return set([])

    @abc.abstractmethod
    def conjugate(self):
        """Complex conjugate"""
        pass

    @abc.abstractmethod
    def _key(self):
        pass


class Variable(Scalar):
    def __init__(self, variable, conjugate=False):
        """Represents a number as a variable

        Parameters
        ----------
        variable : str
            Name of variable.
        conjugate (optional) : bool
            If the variable is conjugated (default: `False`)
        """
        assert_str(variable)
        self._variable = variable
        self._conjugate = conjugate

    def __str__(self):
        if self._conjugate:
            return f"({self._variable}*)"
        else:
            return self._variable

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._variable)}, {self._conjugate})"

    def __mul__(self, other):
        if isinstance(other, Variable):
            if self == other.conjugate():
                return AbsoluteVariable(self._variable)
        return super().__mul__(other)

    def conjugate(self):
        return Variable(self._variable, not self._conjugate)

    def is_zero(self):
        return False

    def is_one(self):
        return False

    def _key(self):
        return (self._variable, self._conjugate)


class AbsoluteVariable(Scalar):
    def __init__(self, variable):
        """Represents the absolute value squared of a variable

        Parameters
        ----------
        variable : str
            Name of variable.
        """
        assert_str(variable)
        self._variable = variable

    def __str__(self):
        return f"|{self._variable}|^2"

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._variable)})"

    def conjugate(self):
        return AbsoluteVariable(self._variable)

    def is_zero(self):
        return False

    def is_one(self):
        return False

    def _key(self):
        return self._variable


class SingleVarFunctionScalar(Scalar):
    def __init__(self, func_name, variable, conjugate=False):
        """Represents a function with a single symbolic variable, e.g. f(x).

        Parameters
        ----------
        func_name : str
            Name of function.
        variable : str
            Name of variable.
        conjugate (optional) : bool
            If the functions is conjugated (default: `False`)
        """
        assert_str(func_name)
        assert_str(variable)
        self._func_name = func_name
        self._variable = variable
        self._conjugate = conjugate

    def __str__(self):
        conj = "*" if self._conjugate else ""
        return f"{self._func_name}{conj}({self._variable})"

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._func_name)}, {repr(self._variable)}, {self._conjugate})"

    def conjugate(self):
        return SingleVarFunctionScalar(self._func_name, self._variable, not self._conjugate)

    def replace_var(self, old_variable, new_variable):
        var = self._variable
        if old_variable == var:
            var = new_variable
        return self.__class__(self._func_name, var, conjugate=self._conjugate)

    def get_variables(self):
        return set([self._variable])

    def is_zero(self):
        return False

    def is_one(self):
        return False

    def has_variable(self, variable):
        return variable == self._variable

    def _key(self):
        return (self._func_name, self._variable, self._conjugate)


class InnerProductFunction(Scalar):
    def __init__(self, func_name1, func_name2):
        """Represents the inner product of two functions.

        Parameters
        ----------
        func_name1 : str
            Name of first function
        func_name2 : str
            Name of second function
        """
        assert_str(func_name1)
        assert_str(func_name2)
        self._func_names = sorted([func_name1, func_name2])

    def __str__(self):
        return f"<{self._func_names[0]}|{self._func_names[1]}>"

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._func_names[0])}, {repr(self._func_names[1])})"

    def conjugate(self):
        return self.__class__(self._func_names[0], self._func_names[1])

    def is_zero(self):
        return False

    def is_one(self):
        return False

    def _key(self):
        return tuple(self._func_names)


class DeltaFunction(Scalar):
    def __init__(self, var1, var2):
        """Delta function between two variables, e.g. d(x - y)

        Parameters
        ----------
        var1 : str
            First variable
        var2 : str
            First variable
        """
        assert_str(var1)
        assert_str(var2)
        self._assert_different(var1, var2)
        self._vars = [var1, var2]

    def conjugate(self):
        return DeltaFunction(*self._vars)

    def __str__(self):
        v1, v2 = self._vars
        return f"D[{v1}-{v2}]"

    def __repr__(self):
        v1, v2 = self._vars
        return f"{self.__class__.__name__}({repr(v1)}, {repr(v2)})"

    def replace_var(self, old_variable, new_variable):
        new_vars = list(self._vars)
        if old_variable in new_vars:
            new_vars.remove(old_variable)
            new_vars.append(new_variable)
        self._assert_different(*new_vars)
        return self.__class__(*new_vars)

    def get_variables(self):
        return set(self._vars)

    def is_zero(self):
        return False

    def is_one(self):
        return False

    def has_variable(self, variable):
        return variable in self._vars

    def _key(self):
        return frozenset(self._vars)

    @staticmethod
    def _assert_different(var1, var2):
        if var1 == var2:
            raise ValueError(f"Variables in a delta function needs to be different, not {var1} and {var2}")


class ProductOfScalars(Scalar):
    def __init__(self, scalars=None):
        """Product of some number of scalars.

        Parameters
        ----------
        scalars : None or list of :class:`~.scalar`
            The factors of the product.
        """
        self._factors = [1]
        if scalars is None:
            return
        assert_list_or_tuple(scalars)
        for scalar in scalars:
            assert_is_scalar(scalar)
            if is_number(scalar):
                self._factors[0] *= scalar
            else:
                self._factors.append(scalar)

    def __str__(self):
        to_print = ""
        for scalar in self:
            to_print += f"{scalar}*"
        return to_print[:-1]

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(list(iter(self)))})"

    def __mul__(self, other):
        if not is_scalar(other):
            return NotImplemented
        if is_number(other):
            return ProductOfScalars([self._factors[0] * other] + self._factors[1:])
        if isinstance(other, ProductOfScalars):
            new_scalar = copy(self)
            for factor in other:
                new_scalar *= factor
            return new_scalar
        # Check if the multiplication between other and any factor can be simplified
        for i, factor in enumerate(self._factors):
            if i == 0:
                continue
            prod = factor * other
            if not isinstance(prod, ProductOfScalars):
                return ProductOfScalars([prod] + self._factors[:i] + self._factors[i + 1:])
        return ProductOfScalars(self._factors + [other])

    def _start_index(self):
        return 1 if (self._factors[0] == 1 and len(self._factors) > 1) else 0

    def __len__(self):
        start = self._start_index()
        return len(self._factors[start:])

    def __iter__(self):
        start = self._start_index()
        return iter(self._factors[start:])

    def __getitem__(self, i):
        start = self._start_index()
        return self._factors[start + i]

    def __setitem__(self, i, value):
        start = self._start_index()
        self._factors[start + i] = value

    def conjugate(self):
        return ProductOfScalars([scalar.conjugate() for scalar in self])

    def atoms(self):
        atoms = []
        for s in self:
            if _is_sequenced_scalar(s):
                atoms += s.atoms()
            else:
                atoms.append(s)
        return atoms

    def replace_var(self, old_variable, new_variable):
        new_factors = []
        for s in self:
            new_factors.append(replace_var(s, old_variable, new_variable))
        return self.__class__(new_factors)

    def get_variables(self):
        vars = set([])
        for factor in self:
            vars |= get_variables(factor)

        return vars

    def expand(self):
        if not any(isinstance(s, SumOfScalars) for s in self._factors):
            # No factor is a sum
            return ProductOfScalars(list(iter(self)))
        expandable_factors = []
        for s in self:
            if isinstance(s, SumOfScalars):
                expandable_factors.append(s._terms)
            else:
                expandable_factors.append([s])

        # Expand the product
        return SumOfScalars([ProductOfScalars(term) for term in product(*expandable_factors)])

    def is_zero(self):
        for f in self._factors:
            if isinstance(f, float):
                if math.isclose(f, 0, abs_tol=1e-16):
                    return True
            else:
                if is_zero(f):
                    return True
        return False

    def is_one(self):
        return all(is_one(s) for s in self._factors)

    def simplify(self, full=True):
        new_scalar = 1
        for factor in self:
            if is_zero(factor):
                return 0
            if not is_one(factor):
                new_scalar *= simplify(factor)
        new_scalar = expand(new_scalar)
        if isinstance(new_scalar, SumOfScalars):
            return simplify(new_scalar)

        if _is_sequenced_scalar(new_scalar):
            if len(new_scalar) == 0:
                return 0
            if len(new_scalar) == 1:
                return simplify(new_scalar[0])

        if isinstance(new_scalar, ProductOfScalars):
            if any(isinstance(s, ProductOfScalars) for s in new_scalar):
                raise RuntimeError("factors of product should not be products")
            if full:
                changed = True
                while changed:
                    changed = new_scalar._combine_factors()
        return new_scalar

    def _combine_factors(self):
        """Test if factors can be combined"""
        for i, factor1 in enumerate(self._factors):
            if i == 0:
                continue
            if is_number(factor1):
                self._factors[0] += factor1
                self._factors.pop(i)
                return True
            for j in range(i + 1, len(self._factors)):
                factor2 = self._factors[j]
                prod = factor1 * factor2
                if not isinstance(prod, ProductOfScalars) or len(prod._factors) <= 2:
                    self._factors[i] = prod
                    self._factors.pop(j)
                    return True
        return False

    def has_variable(self, variable):
        return any(has_variable(factor, variable) for factor in self)

    def _key(self):
        factors_with_multi = defaultdict(int)
        for factor in self:
            factors_with_multi[factor] += 1
        return frozenset(factors_with_multi.items())


class SumOfScalars(Scalar):
    def __init__(self, scalars=None):
        """Sum of some number of scalars.

        Parameters
        ----------
        scalars : None or list of :class:`~.scalar`
            The terms of the sum.
        """
        self._terms = [0]
        if scalars is None:
            return
        assert_list_or_tuple(scalars)
        for scalar in scalars:
            assert_is_scalar(scalar)
            if is_number(scalar):
                self._terms[0] += scalar
            else:
                self._terms.append(scalar)

    def __str__(self):
        to_print = ""
        for scalar in self:
            to_print += f"{scalar} + "
        return "(" + to_print[:-3] + ")"

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(list(iter(self._terms)))})"

    def __add__(self, other):
        # Check if one is zero
        if is_zero(other):
            return copy(self)
        if is_zero(self):
            return copy(other)

        if not is_scalar(other):
            return NotImplemented
        if is_number(other):
            return SumOfScalars([self._terms[0] + other] + self._terms[1:])
        if isinstance(other, SumOfScalars):
            new_scalar = copy(self)
            for term in other:
                new_scalar += term
            return new_scalar
        # Check if the addition between other and any term can be simplified
        for i, term in enumerate(self._terms):
            if i == 0:
                continue
            sm = term + other
            if not isinstance(sm, SumOfScalars):
                return SumOfScalars([sm] + self._terms[:i] + self._terms[i + 1:])
        return SumOfScalars(self._terms + [other])

    def _start_index(self):
        return 1 if self._terms[0] == 0 else 0

    def __len__(self):
        start = self._start_index()
        return len(self._terms[start:])

    def __iter__(self):
        start = self._start_index()
        return iter(self._terms[start:])

    def __getitem__(self, i):
        start = self._start_index()
        return self._terms[start + i]

    def __setitem__(self, i, value):
        start = self._start_index()
        self._terms[start + i] = value

    def conjugate(self):
        return SumOfScalars([scalar.conjugate() for scalar in self._terms])

    def atoms(self):
        atoms = []
        for s in self:
            if _is_sequenced_scalar(s):
                atoms += s.atoms()
            else:
                atoms.append(s)
        return atoms

    def replace_var(self, old_variable, new_variable):
        new_terms = []
        for s in self:
            new_terms.append(replace_var(s, old_variable, new_variable))
        return self.__class__(new_terms)

    def get_variables(self):
        vars = set([])
        for term in self:
            vars |= get_variables(term)

        return vars

    def is_zero(self):
        return all(is_zero(s) for s in self)

    def expand(self):
        return sum((expand(term) for term in self))

    def simplify(self, full=True):
        new_scalar = 0
        expanded = self.expand()

        if not isinstance(expanded, SumOfScalars):
            return simplify(expanded)
        if is_number(expanded):
            return expanded
        for term in expanded:
            if not is_zero(term):
                new_scalar += simplify(term)

        if _is_sequenced_scalar(new_scalar):
            if len(new_scalar) == 0:
                new_scalar = 0
            if len(new_scalar) == 1:
                new_scalar = simplify(new_scalar[0])

        if isinstance(new_scalar, SumOfScalars):
            if any(isinstance(s, SumOfScalars) for s in new_scalar):
                raise RuntimeError("terms of sum should not be sum after expand")
            if full:
                changed = True
                while changed:
                    changed = new_scalar._combine_terms()
        return new_scalar

    def _combine_terms(self):
        """Test if terms can be combined"""
        for i, term1 in enumerate(self._terms):
            if i == 0:
                continue
            if is_number(term1):
                self._terms[0] += term1
                self._terms.pop(i)
                return True
            for j in range(i + 1, len(self._terms)):
                term2 = self._terms[j]
                sm = term1 + term2
                if not isinstance(sm, SumOfScalars) or len(sm._terms) <= 2:
                    self._terms[i] = sm
                    self._terms.pop(j)
                    return True
        return False

    def has_variable(self, variable):
        return any(has_variable(factor, variable) for factor in self)

    def _key(self):
        terms_with_multi = defaultdict(int)
        for term in self:
            terms_with_multi[term] += 1
        return frozenset(terms_with_multi.items())


def assert_is_scalar(n):
    """Asserts that something is a scalar"""
    if not is_scalar(n):
        raise TypeError(f"variable ({n}) should be a scalar, not {type(n)}")


def _get_multiple_of_scalar(scalar):
    if isinstance(scalar, ProductOfScalars):
        if len(scalar._factors) == 2:
            return tuple(scalar._factors)
        else:
            return scalar._factors[0], ProductOfScalars(scalar._factors[1:])
    return 1, scalar


def _is_sequenced_scalar(scalar):
    return any(isinstance(scalar, tp) for tp in [ProductOfScalars, SumOfScalars])



"""
Module for performing integration of scalars (:func:`~.scalar.Scalar`).

Main function is :func:`~.integrate` which takes a scalar and the variables to integrate over.
"""
from copy import copy

from qualg.toolbox import assert_str, replace_var, simplify, get_variables, has_variable
from qualg.scalars import is_number, DeltaFunction, SumOfScalars, ProductOfScalars,\
    InnerProductFunction, SingleVarFunctionScalar, Scalar, assert_is_scalar


class _Integration(Scalar):

    def __init__(self, scalar, variable):
        assert_is_scalar(scalar)
        assert_str(variable)
        if not isinstance(scalar, ProductOfScalars):
            raise TypeError(f"scalar should be ProductOfScalars, not {type(scalar)}")
        if not all(has_variable(factor, variable) for factor in scalar):
            raise ValueError("all factors should have the integration term")
        self._scalar = scalar
        self._variable = variable

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        # TODO check with swapping the integration variable since this shouldn't matter
        return self._scalar == other._scalar and self._variable == other._variable

    def __hash__(self):
        return hash(self._key())

    def __str__(self):
        return f"S_{self._variable}{{{self._scalar}}}"

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._scalar)}, {repr(self._variable)}"

    def __copy__(self):
        return self.__class__(self._scalar, self._variable)

    def conjugate(self):
        return self.__class__(self._scalar.conjugate(), self._variable)

    def simplify(self):
        new_scalar = copy(self)
        for evaluation in EVALUATIONS:
            new_scalar = evaluation(new_scalar)
            if not isinstance(new_scalar, _Integration):
                return new_scalar
        return new_scalar

    def has_variable(self, variable):
        # raise NotImplementedError()
        if variable == self._variable:
            return False
        return has_variable(self._scalar, variable)

    def _key(self):
        return (self._scalar, self._variable)


def integrate(scalar, variable=None):
    """
    Integrates a scalar over a given variable or variables.

    Parameters
    ----------
    scalar : :class:`~.scalar.Scalar`
        The scalar to integrate.
    variable : None or str or set of str
        The variable(s) to integrate over.
        Can be:
        * `None`: Then all variables in the scalar are integrated out.
        * `str`: Then a single variable is integrated out.
        * `set` of `str`: Then all the variables in the set are integrated out.

    Returns
    -------
    :class:`~.scalar.Scalar`
        The output of the integration.
    """
    # TODO needed?
    scalar = simplify(scalar)
    if variable is None:
        return integrate(scalar, get_variables(scalar))
    if isinstance(variable, set):
        new_scalar = scalar
        for v in variable:
            new_scalar = integrate(new_scalar, v)
        return new_scalar
    assert_str(variable)
    if isinstance(scalar, SumOfScalars):
        new_scalar = sum(integrate(s, variable) for s in scalar._terms)
    # elif isinstance(scalar, ProductOfScalars):
    elif isinstance(scalar, Scalar):
        if isinstance(scalar, ProductOfScalars):
            factors = scalar._factors
        else:
            factors = [scalar]
        # Split factors based on if they contain the integration variable or not
        var_factors = []
        other_factors = []
        for factor in factors:
            if isinstance(factor, Scalar) and has_variable(factor, variable):
                var_factors.append(factor)
            else:
                other_factors.append(factor)
        if len(var_factors) > 0:
            integration_part = [_Integration(ProductOfScalars(var_factors), variable)]
        else:
            integration_part = []

        new_scalar = ProductOfScalars(other_factors + integration_part)
    elif is_number(scalar):
        new_scalar = scalar
    else:
        raise NotImplementedError(f"integrate not implemented for type {type(scalar)}")

    return simplify(new_scalar)


def _evaluate_delta_function(integration_scalar):
    integrand = integration_scalar._scalar
    variable = integration_scalar._variable
    # Find the delta functions containing this variable
    deltas = [(i, d) for (i, d) in enumerate(integrand._factors)
              if isinstance(d, DeltaFunction) and variable in d._vars]
    if len(deltas) == 0:
        return integration_scalar
    # Replace variables with the variable to the other variable in the delta function
    i, delta = deltas[0]
    # Get the other variable in the delta function
    try:
        other_var = next(v for v in delta._vars if v != variable)
    except StopIteration:
        # TODO This should not happen anymore
        raise RuntimeError(f"Encountered delta function with the same variable: {delta}")
    # Replace the delta function with 1 (recall that replace_var creates a copy)
    integrand._factors[i] = 1
    integrand = replace_var(integrand, old_variable=variable, new_variable=other_var)

    return integrand


def _find_norm_identities(integration_scalar):
    """Finds integrals which are the norm of a function, i.e. 1"""
    integrand = integration_scalar._scalar
    # TODO generalise this
    if len(integrand) == 2:
        if all(isinstance(s, SingleVarFunctionScalar) for s in integrand):
            if integrand[0] == integrand[1].conjugate():
                return 1
    return integration_scalar


def _find_function_inner_products(integration_scalar):
    """Finds integrals which evaluate to the inner product of functions."""
    integrand = integration_scalar._scalar
    # TODO generalise this
    if len(integrand) == 2:
        if all(isinstance(s, SingleVarFunctionScalar) for s in integrand):
            return InnerProductFunction(integrand[0]._func_name, integrand[1]._func_name)
    return integration_scalar


# These evaluations are used when integrating
EVALUATIONS = [
    _evaluate_delta_function,
    _find_norm_identities,
    _find_function_inner_products,
]



"""
Contains classes for representing in excited states of some bosonic mode.
The creation/annihilation operators have a symbolic variable which can for example
be the frequency of the excited state.
"""
from collections import defaultdict
from itertools import permutations

from qualg.scalars import DeltaFunction
from qualg.states import BaseState
from qualg.toolbox import assert_str, assert_list_or_tuple, replace_var


class FockOp:
    def __init__(self, mode, variable, creation=True):
        """
        Represents an creation/annihilation operator in a given mode with a given variable
        representing for example the frequency of the excitation.

        Parameters
        ----------

        mode: str
            The mode of the excitation
        variable: str
            The variable describing the continous variable
        creation : bool
            Whether this is a creation operator or not (annihilation).
        """
        assert_str(mode)
        assert_str(variable)
        self._mode = mode
        self._variable = variable
        self._creation = creation

    def _key(self):
        return self._mode, self._variable, self._creation

    def __eq__(self, other):
        if not isinstance(other, FockOp):
            return NotImplemented
        return self._key() == other._key()

    def __hash__(self):
        return hash(self._key())

    def __str__(self):
        dag = "+" if self._creation else ""
        return f"{self._mode}{dag}({self._variable})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self._mode}, {self._variable}, {self._creation})"

    def dagger(self):
        """
        Complex conjugate of the operator.
        """
        return FockOp(self._mode, self._variable, not self._creation)

    def replace_var(self, old_variable, new_variable):
        """
        Replaces a variable with another.
        """
        var = self._variable
        if old_variable == var:
            var = new_variable
        return self.__class__(self._mode, var, creation=self._creation)

    def get_variables(self):
        """
        Returns the variable of this operator.
        """
        return set([self._variable])


class FockOpProduct:
    def __init__(self, fock_ops=None):
        """
        A product of excitation operators (:class:`~.FockOp`).

        Parameters
        ----------
        fock_ops : list of :class:`~.FockOp`
            The product of fock operators.
        """
        self._fock_ops = defaultdict(int)
        if fock_ops is None:
            return
        assert_list_or_tuple(fock_ops)
        for fock_op in fock_ops:
            assert_fock_op(fock_op)
            # TODO we only allow these to be creation ops for now
            if not fock_op._creation:
                raise NotImplementedError
            self._fock_ops[fock_op] += 1

    def __mul__(self, other):
        new_op = FockOpProduct()
        if isinstance(other, FockOp):
            for fock_op, count in self._fock_ops.items():
                new_op._fock_ops[fock_op] += count
            new_op._fock_ops[other] += 1
            return new_op
        elif isinstance(other, FockOpProduct):
            for fock_op, count in self._fock_ops.items():
                new_op._fock_ops[fock_op] += count
            for fock_op, count in other._fock_ops.items():
                new_op._fock_ops[fock_op] += count

        return new_op

    def __str__(self):
        to_print = ""
        for fock_op, count in self._fock_ops.items():
            to_print += f"{fock_op}^{count} * "
        return to_print[:-3]

    def _key(self):
        return tuple(sorted(self._fock_ops.items(), key=lambda x: x[0]._key()))

    def __hash__(self):
        return hash(self._key())

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._key() == other._key()

    def dagger(self):
        """
        Complex conjugate of the operator.
        """
        new_op = FockOpProduct()
        for fock_op, count in self._fock_ops.items():
            new_op._fock_ops[fock_op.dagger()] = count

        return new_op

    def variables_in_mode(self, mode):
        """
        Get the variables in a given mode.
        """
        variables = []
        for fock_op, count in self._fock_ops.items():
            if fock_op._mode == mode:
                for _ in range(count):
                    variables.append(fock_op._variable)

        return variables

    def variables_by_modes(self):
        """
        Get a dictionay of variables per modes.
        """
        variables = defaultdict(list)
        for fock_op, count in self._fock_ops.items():
            mode = fock_op._mode
            for _ in range(count):
                variables[mode].append(fock_op._variable)

        return variables

    def replace_var(self, old_variable, new_variable):
        """
        Replaces a variable with another.
        """
        new_fock_op_product = FockOpProduct()
        for fock_op, count in self._fock_ops.items():
            new_fock_op_product._fock_ops[replace_var(fock_op, old_variable, new_variable)] = count
        return new_fock_op_product

    def get_variables(self):
        """
        Returns the variable of this operator.
        """
        vars = set([])
        for fock_op in self._fock_ops:
            vars |= fock_op.get_variables()

        return vars


class BaseFockState(BaseState):
    def __init__(self, fock_ops=None):
        """
        A base state represented by excitations from vacuum.

        Parameters
        ----------
        fock_ops : :class:`.FockOpProduct` or list of :class:`~.FockOp`
            The product of fock operators.
        """
        if isinstance(fock_ops, FockOpProduct):
            self._fock_op_product = fock_ops
        else:
            self._fock_op_product = FockOpProduct(fock_ops)

    def _key(self):
        return self._fock_op_product._key()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._key() == other._key()

    def __hash__(self):
        return hash(self._key())

    def __str__(self):
        to_print = ""
        for fock_op, count in self._fock_op_product._fock_ops.items():
            to_print += f"{fock_op}^{count}"
        return to_print + "|0>"

    @property
    def shape(self):
        return None

    def inner_product(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()
        l_vars_by_mode = self._fock_op_product.variables_by_modes()
        r_vars_by_mode = other._fock_op_product.variables_by_modes()
        all_modes = set(l_vars_by_mode.keys()) | set(r_vars_by_mode.keys())
        scalar = 1
        for mode in all_modes:
            l_vars = l_vars_by_mode.get(mode, [])
            r_vars = r_vars_by_mode.get(mode, [])
            if len(l_vars) != len(r_vars):
                return 0
            factor = 0
            for perm_r_vars in permutations(r_vars):
                term = 1
                for l_var, r_var in zip(l_vars, perm_r_vars):
                    term *= DeltaFunction(l_var, r_var)
                factor += term
            scalar *= factor

        return scalar

    def tensor_product(self, other):
        if not isinstance(other, self.__class__):
            raise NotImplementedError(f"fock tensor product is not implemented for {type(other)}")
        prod = self._fock_op_product * other._fock_op_product
        return BaseFockState(fock_ops=prod)

    def replace_var(self, old_variable, new_variable):
        """
        Replaces a variable with another.
        """
        return self.__class__(replace_var(self._fock_op_product, old_variable, new_variable))

    def get_variables(self):
        """
        Returns the variable of this operator.
        """
        return self._fock_op_product.get_variables()

    def _compatible(self, other):
        if not isinstance(other, BaseFockState):
            return False
        return True

    def _bra_str(self):
        to_print = ""
        for fock_op, count in self._fock_op_product._fock_ops.items():
            to_print += f"{fock_op.dagger()}^{count}"
        return "<0|" + to_print


def assert_fock_op(op):
    """
    Asserts that an object is a fock operator.
    """
    if not isinstance(op, FockOp):
        raise TypeError(f"operator should be of type FockOp, not {type(op)}")
